#include <iostream>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>
#include "kapi.h"

using namespace std;

int usage(char *basename)
{
    cerr << "Usage: " << basename << " hostname portnum module function\n";
    return (-1);
}

main(int argc, char **argv)
{
    kapi_manager kmgr;

    if (argc != 5) {
	return usage(argv[0]);
    }

    const char *host = argv[1];
    int port = atoi(argv[2]);
    const char *modName = argv[3];
    const char *funcName = argv[4];

    // Attach to the kernel on the target host
    cout << "Attaching to the kernel on host " << host
	 << " portnum " << port << "...\n";
    if (kmgr.attach(host, port) < 0) {
	abort();
    }

    // Find the module of interest, initialize kmod
    kapi_module kmod;
    if (kmgr.findModule(modName, &kmod) < 0) {
	abort();
    }

    // Find the function of interest, initialize kfunc
    kapi_function kfunc;
    if (kmod.findFunction(funcName, &kfunc) < 0) {
	abort();
    }

    // Find the instrumentation point: the entry to the function
    kapi_vector<kapi_point> entries;
    if (kfunc.findEntryPoint(&entries) < 0) {
	abort();
    }

    // Allocate space for the counter in the kernel space
    kptr_t addr;
    unsigned size = sizeof(kapi_int_t);
    if ((addr = kmgr.malloc(size)) == 0) {
	abort();
    }

    // Initialize the kernel memory, by copying buffer from the user space
    kapi_int_t user_copy = 0;
    if (kmgr.memcopy(&user_copy, addr, size) < 0) {
	abort();
    }

    // Create an int variable in the allocated space
    kapi_int_variable intCounter(addr);
    
    // Generate code to increment the variable atomically
    kapi_arith_expr code(kapi_atomic_assign, intCounter, 
			 kapi_arith_expr(kapi_plus, intCounter, 
					 kapi_const_expr(1)));

    // Splice the code at specified points
    int sh;
    if ((sh = kmgr.insertSnippet(code, entries[0])) < 0) {
	abort();
    }

    // Loop for some time
    time_t t = time(NULL);
    while (time(NULL) - t < 20) {
	kmgr.handleEvents();
	if (kmgr.memcopy(addr, &user_copy, size) < 0) {
	    abort();
	}
	// The client's byte ordering may not match that of the kernel
	kmgr.to_client_byte_order(&user_copy, size);

	cout << "Counter = " << user_copy << endl;
	sleep(1);
    }

    // Clean-up
    if (kmgr.removeSnippet(sh) < 0) {
	abort();
    }
    
    // Free the kernel memory
    if (kmgr.free(addr) < 0) {
	abort();
    }
    
    // Detach from the kernel
    if (kmgr.detach() < 0) {
	abort();
    }
    cout << "Detached from the kernel\n";
}

